% Diverse Sortieralgorithmen, wobei man von asort/3 alle aufrufen kann


% Beispiel Anfrage: liste(X), asort(quicksort, X, Sortiert)
asort(Art, Liste, Sortiert) :- Proz =.. [Art, Liste, Sortiert], call(Proz).
liste([15,25,3,66,8]).


% Hilfsprdikat, fgt die zweite Liste an die erste
append([], Liste, Liste).
append([Kopf | Rest], Liste, [Kopf | Rest2]) :- append(Rest, Liste, Rest2).


% Sortierung durch Permutation
gruselsort(Liste, Sortiert) :- permutation(Liste, Sortiert),
                               sortiert(Sortiert).                    
permutation([], []).
permutation(Liste, [Kopf | Rest]) :- append(L1, [Kopf | L2], Liste),
                                     append(L1, L2, L3),
                                     permutation(L3, Rest).                   
sortiert([]).
sortiert([_]).
sortiert([Elem1, Elem2 | Rest]) :- Elem1 =< Elem2,
                                   sortiert([Elem2 | Rest]).
                                   
% Tauschsort                                   
tauschsort(Liste, Sortiert) :- append(Anfang, [E1, E2 | Ende], Liste),
                               not(relation(E1, E2)), !,
                               append(Anfang, [E2, E1 | Ende], Zwi),
                               tauschsort(Zwi, Sortiert).
tauschsort(Liste, Liste).

% Quicksort
quicksort([], []) :- !.
quicksort([Elem | Rest], Sortiert) :- partition(Elem, Rest, Kleinere, Groessere),
                                      quicksort(Kleinere, Kleinsortiert),
                                      quicksort(Groessere, Grosssortiert),
                                      append(Kleinsortiert, [Elem | Grosssortiert], Sortiert).
               
% Quicksort (verbessert)
quicker(Liste, Sortiert) :- quicker1(Liste, Sortiert, []).
quicker1([], Liste, Liste) :- !.
quicker1([Elem | Liste], Sort1, Ende) :- partition(Elem, Liste, Kleiner, Groesser),
                                         quicker1(Kleiner, Sort1, [Elem | Rest]),
                                         quicker1(Groesser, Rest, Ende).                     
partition(_, [], [], []) :- !.
partition(Pivot, [Elem | Rest], [Elem | KleinRest], GrossRest) :-
                                Elem =< Pivot,
                                partition(Pivot, Rest, KleinRest, GrossRest).
partition(Pivot, [Elem | Rest], KleinRest, [Elem | GrossRest]) :-
                                partition(Pivot, Rest, KleinRest, GrossRest).
                                
% Mergesort                               
mergesort(Liste, Sortiert) :- l_von_l(Liste, L_von_L),
                              msort2(L_von_L, [Sortiert]).                           
msort2([], []).
msort2([Liste], [Liste]).
msort2([Liste1, Liste2 | Rest], Sortiert) :- merge2(Liste1, Liste2, Merged),
                                            msort2(Rest, Sortiert1),
                                            msort2([Merged | Sortiert1], Sortiert).
l_von_l([], []).
l_von_l([Element | Restliste], [[Element] | Restliste2]) :-
                                          l_von_l(Restliste, Restliste2).
merge2([], L, L).
merge2(L, [], L).
merge2([Elem1 | Rest1], [Elem2 | Rest2], [Elem1 | Merged]) :-
                                Elem1 =< Elem2,
                                merge2(Rest1, [Elem2 | Rest2], Merged).
merge2([Elem1 | Rest1], [Elem2 | Rest2], [Elem2 | Merged]) :-
                                merge2([Elem1 | Rest1], Rest2, Merged).
                                
% Treesort                                
treesort(Liste, Sortiert) :- tsort(Liste, Baum), blaetter(Baum, Sortiert).
tsort([], _).
tsort([Elem | Rest], Baum) :- find(Elem, Baum), tsort(Rest, Baum).
find(E, tree(E,_,_)).
find(E, tree(E1, Kleiner, _)) :- E \= E1, E =< E1, find(E, Kleiner).
find(E, tree(E1, _, Groesser)) :- E \= E1, find(E, Groesser).
blaetter(Baum, []) :- var(Baum), !.
blaetter(tree(Knoten, Links, Rechts), Liste) :- blaetter(Links, LListe),
                                                blaetter(Rechts, RListe),
                                                append(LListe, [Knoten | RListe], Liste).
